
/***************************************************************************
 *   This file is part of Aspect, a simple PEC tool.                       *
 *                                                                         *
 *   Copyright (C) 2006-2007 by Wolfgang Hoffmann <woho@woho.de>           *
 *                                                                         *
 *   This program is free software, licensed under the GPL v2.             *
 *   See the file COPYING for more details.                                *
 ***************************************************************************/


#include "datapemcu.h"
#include "data.h"
#include "log.h"

#include <QtGui>
#include <math.h>


DataPeMcu::DataPeMcu()
:   m_bInModelRefill(false)
    {
    m_qpModel = new QStandardItemModel(this);

    m_aPeProcessed.resize(PecData::c_nRowNum);
    m_aPecIdeal.resize(PecData::c_nRowNum);

    connect(m_qpModel, SIGNAL(itemChanged(QStandardItem *)), this, SLOT(changePeItem(QStandardItem *)));
    connect(&m_oPoller, SIGNAL(timeout()), this, SLOT(readPeFile()));
    }


DataPeMcu::~DataPeMcu()
    {
    if (m_oPeFile.isOpen())
        m_oPeFile.close();
    if (!m_qpModel.isNull())
        delete (QStandardItemModel *)m_qpModel;
    }


void DataPeMcu::clearData()
    {
    pollPeFile(false);
    clearPeData();
    }


bool DataPeMcu::probePeFile(const QString &qsFileName) const
    {
    LOG(4, "DataPeMcu::probePeFile(%s)", LQS(qsFileName));

    bool bIsMcuFormat = false;
    QFile oPeFile(qsFileName);
    if (oPeFile.open(QIODevice::ReadOnly | QIODevice::Text))
        {
        QString qsHeader = QString::fromLocal8Bit(oPeFile.readLine()).trimmed();
        LOG(5, "    %s", LQS(qsHeader));
        if (qsHeader == QString::fromUtf8("Time, PEC Table, RA Error, DEC Error"))
            bIsMcuFormat = true;
        oPeFile.close();
        }

    return bIsMcuFormat;
    }


void DataPeMcu::setPeFileName(const QString &qsFileName)
    {
    if (m_oPeFile.isOpen())
        m_oPeFile.close();
    m_oPeFile.setFileName(qsFileName);
    }


bool DataPeMcu::loadPeFile()
    {
    LOG(4, "DataPeMcu::loadPeFile(%s)", LQS(m_oPeFile.fileName()));

    if (m_oPeFile.isOpen())
        m_oPeFile.close();

    clearData();
    if (!openPeFile())
        return false;
    readPeFile();
    closePeFile();

    return true;
    }


bool DataPeMcu::pollPeFile(bool bOn)
    {
    if (bOn)
        {
        // verify that polling was off
        if (m_oPoller.isActive())
            return false;
        clearData();
        if (!openPeFile())
            return false;
        m_oPoller.start(1000);
        m_qsLineFragment = QString();
        readPeFile();
        }
    else
        {
        // verify that polling was on
        if (!m_oPoller.isActive())
            return false;
        m_oPoller.stop();
        closePeFile();
        }
    return true;
    }


/*!
    Clear data list m_lstPeRaw and model m_qpModel.

    Open PE data file, asuming that filename on m_oPeFile is already
    set. Set device to text stream and read and check file header.
*/

bool DataPeMcu::openPeFile()
    {
    if (!m_oPeFile.open(QIODevice::ReadOnly | QIODevice::Text))
        {
        LOG(0, "DataPeMcu::loadPeFile(%s): failed to open file", LQS(m_oPeFile.fileName()));
        return false;
        }
    QString qsHeader = QString::fromLocal8Bit(m_oPeFile.readLine()).trimmed();
    if (qsHeader != QString::fromUtf8("Time, PEC Table, RA Error, DEC Error"))
        {
        LOG(0, "DataPeMcu::loadPeFile(%s): wrong file format:", LQS(m_oPeFile.fileName()));
        LOG(0, "    \"%s\"", LQS(qsHeader));
        return false;
        }

    return true;
    }


void DataPeMcu::closePeFile()
    {
    if (!m_oPeFile.isOpen())
        return;
    m_oPeFile.close();
    }


/*!
    Clear m_lstPeRaw and clear and reset model m_qpModel
*/

void DataPeMcu::clearPeData()
    {
    m_qpModel->clear();

    m_lstPeRaw.clear();
    m_qpModel->setColumnCount(5);
    QStringList qslHeaders;
    qslHeaders << trUtf8("Time");
    qslHeaders << trUtf8("Worm Position");
    qslHeaders << trUtf8("RA Error");
    qslHeaders << trUtf8("DEC Error");
    qslHeaders << trUtf8("");
    m_qpModel->setHorizontalHeaderLabels(qslHeaders);
    }


/*!
    Read lines from file, as long as complete lines of data
    are available.

    Enter data into m_lstPeRaw while reading, and update
    model m_qpModel after reading is done. Finally emit
    dataPeChanged().

    Don't use QTextStream, as it obviously doesn't update
    it's internal buffer when data is appended to a file.
*/

void DataPeMcu::readPeFile()
    {
    LOG(5, "readPeFile()");
    while (m_oPeFile.bytesAvailable() > 0)
        {
        QByteArray baLine = m_oPeFile.readLine();
        // careful:
        // if QByteArray is empty, no more complete lines available.
        // otherwise if QString is empty, skip empty line
        if (baLine.isEmpty())
            break;
        // need to keep a simple buffer for keeping incomplete lines accross readPeFile() calls
        m_qsLineFragment += QString::fromLocal8Bit(baLine).trimmed();
        if (!baLine.endsWith('\n'))
            break;
        QString qsLine = m_qsLineFragment;
        m_qsLineFragment = QString();
        LOG(5, "    \"%s\"", LQS(qsLine));
        if (qsLine.isEmpty())
            continue;
        QStringList qslEntryStrings = qsLine.split(",");
        PeRawSample oSample;
        oSample.m_oTime = QTime();
        if (qslEntryStrings.size() > 4)
            {
            // cope with files written under German language settings,
            // where decimal separators '.' becomes ','
            // heuristically convert every second ',' back to '.',
            // starting from backwards, as timestamp might have no decimals
            int nNum = 0;
            for (int nPos = qsLine.lastIndexOf(','); nPos > 0; nPos = qsLine.lastIndexOf(',', nPos-1))
                {
                if ((nNum % 2) == 0)
                    qsLine.replace(nPos, 1, '.');
                nNum++;
                }
            qslEntryStrings = qsLine.split(",");
            }
        if (qslEntryStrings.size() > 0)
            {
            oSample.m_oTime = QTime::fromString(qslEntryStrings.at(0), QString::fromUtf8("hh:mm:ss.zzz"));
            if (!oSample.m_oTime.isValid())
                oSample.m_oTime = QTime::fromString(qslEntryStrings.at(0), QString::fromUtf8("hh:mm:ss"));
            }
        oSample.m_dfWorm = (qslEntryStrings.size() > 1)? qslEntryStrings.at(1).toDouble(): 0.;
        oSample.m_dfRa = (qslEntryStrings.size() > 2)? qslEntryStrings.at(2).toDouble(): 0.;
        oSample.m_dfDec = (qslEntryStrings.size() > 3)? qslEntryStrings.at(3).toDouble(): 0.;
        m_lstPeRaw.append(oSample);
        }

    m_bInModelRefill = true;
    int nFirstRow = m_qpModel->rowCount();
    int nLastRow = m_lstPeRaw.size();
    m_qpModel->setRowCount(nLastRow);
    for (int nRow = nFirstRow; nRow < nLastRow; nRow++)
        {
        QStandardItem *pItem;
        pItem = new QStandardItem(m_lstPeRaw.at(nRow).m_oTime.toString(QString::fromUtf8("hh:mm:ss.zzz")));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_qpModel->setItem(nRow, 0, pItem);
        pItem = new QStandardItem(QString::number(m_lstPeRaw.at(nRow).m_dfWorm, 'f', 3));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_qpModel->setItem(nRow, 1, pItem);
        pItem = new QStandardItem(QString::number(m_lstPeRaw.at(nRow).m_dfRa));
        pItem->setEditable(true);
        pItem->setSelectable(true);
        m_qpModel->setItem(nRow, 2, pItem);
        pItem = new QStandardItem(QString::number(m_lstPeRaw.at(nRow).m_dfDec));
        pItem->setEditable(true);
        pItem->setSelectable(true);
        m_qpModel->setItem(nRow, 3, pItem);
        }
    m_bInModelRefill = false;

    emit dataPeChanged();
    }


void DataPeMcu::changePeItem(QStandardItem *pItem)
    {
    LOG(5, "DataPeMcu::changePeItem");

    if ((pItem == 0) || m_bInModelRefill)
        return;
    int nCol = pItem->column();
    if ((nCol != 2) && (nCol != 3))
        return;
    int nRow = pItem->row();
    if ((nRow < 0) || (nRow >= m_lstPeRaw.size()))
        return;
    double dfValue = pItem->text().toDouble();
    if (nCol == 2)
        {
        if (m_lstPeRaw.at(nRow).m_dfRa == dfValue)
            return;
        LOG(1, "manually changing PE measurement item %d Ra from %0.3f to %0.3f",
            nRow, m_lstPeRaw.at(nRow).m_dfRa, dfValue);
        m_lstPeRaw[nRow].m_dfRa = dfValue;
        }
    else
        {
        if (m_lstPeRaw.at(nRow).m_dfDec == dfValue)
            return;
        LOG(1, "manually changing PE measurement item %d Dec from %0.3f to %0.3f",
            nRow, m_lstPeRaw.at(nRow).m_dfDec, dfValue);
        m_lstPeRaw[nRow].m_dfDec = dfValue;
        }

    emit dataPeChanged();
    }


/*!
    \todo rework drift compensation
        - it relies on the presence of precise timestamps
            (it timestamp lack msecs, worm pos would have better resolution, but is
            proportional to time only if PE measurement was done with PEC and IntellyTrack off)
        - it probably should consider the DEC values and work with something
            more sophisticated than linear RA scaling
*/

void DataPeMcu::processData(int nBeg, int nEnd, bool bAverage, int nSmoothWidth,
    double dfDrift, double dfScale, double dfOffset,
    PecData *pPecTable, int nWormWheelTeeth, double dfPecPosScale)
    {
    bool bTouchup = (pPecTable != 0);

    m_lstPePure.clear();

    if (nBeg < 0)
        nBeg = 0;
    if ((nEnd < 0) || (nEnd > m_lstPeRaw.size()))
        nEnd = m_lstPeRaw.size();
    for (int nRow = nBeg; nRow < nEnd; nRow++)
        {
        PePureSample oSample;
        oSample.m_dfWorm = m_lstPeRaw.at(nRow).m_dfWorm;
        double dfTime = m_lstPeRaw.at(0).m_oTime.msecsTo(m_lstPeRaw.at(nRow).m_oTime);
        oSample.m_dfPe = (m_lstPeRaw.at(nRow).m_dfRa - dfTime * dfDrift) * dfScale + dfOffset;
        m_lstPePure.append(oSample);
        LOG(5, "    %.03f: %.03f", oSample.m_dfWorm, oSample.m_dfPe);
        }

    // fist smooth raw sample curves for display
    m_lstPeSmoothed.clear();
    if ((nSmoothWidth == 0) || (m_lstPePure.isEmpty()))
        {
        // if without smoothing, pass through original data
        m_lstPeSmoothed = m_lstPePure;
        }
    else
        {
        // combine samples into PEC point grid,
        // taking into account samples of nSmoothWidth PEC points

        // walk over raw data, keeping record of begin, mid and end of
        // smoothing interval:
        // - advance end of interval until hitting the end of a PEC point
        // - advance begin of interval as long as it's further from end
        //   than smoothing width
        // - calculate average/median of data samples within the interval
        //   and store at end index into list of smoothed samples
        int nRowBeg = -1;
        int nRowEnd = 0;
        int nPecMid = -1;
        int nPecBeg = -1;
        int nPecEnd = -1;
        for (int nRowMid = 0; nRowMid < m_lstPePure.size(); nRowMid++)
            {
            int nPecPrev = nPecMid;
            nPecMid = (int)floor(m_lstPePure.at(nRowMid).m_dfWorm);
            if (nPecMid == nPecPrev)
                continue;

            // advance begin of data range while data sample is outside
            // left half of smoothing interval
            // be careful to cope with both PEC table traversal directions
            // and correctly deal with PEC point wraparound and initial case
            for (;;)
                {
                int nPecDiff = abs(nPecMid - nPecBeg);
                if (nPecDiff > PecData::c_nRowNum / 2)
                    nPecDiff = abs(PecData::c_nRowNum - nPecDiff);
                if ((nPecBeg >= 0) && (nPecDiff <= nSmoothWidth / 2))
                    break;
                nRowBeg += 1;
                nPecBeg = (int)floor(m_lstPePure.at(nRowBeg).m_dfWorm);
                }
            // dito for last data sample. nRowEnd is exclusive.
            // watch of end of data, and make sure that nRowEnd is above nRowMid
            while (nRowEnd < m_lstPePure.size())
                {
                int nPecDiff = abs(nPecMid - nPecEnd);
                if (nPecDiff > PecData::c_nRowNum / 2)
                    nPecDiff = abs(PecData::c_nRowNum - nPecDiff);
                if ((nPecEnd >= 0) && (nRowEnd > nRowMid) && (nPecDiff > (nSmoothWidth - 1) / 2))
                    break;
                nPecEnd = (int)floor(m_lstPePure.at(nRowEnd).m_dfWorm);
                nRowEnd += 1;
                }
            LOG(5, "    [%d,%d,%d[: %d,%d,%d", nRowBeg, nRowMid, nRowEnd, nPecBeg, nPecMid, nPecEnd);

            PePureSample oSample;
            oSample.m_dfWorm = (double)nPecMid;
            if (nRowEnd - nRowBeg <= 1)
                {
                // one single sample: use directly
                oSample.m_dfPe = m_lstPePure.at(nRowMid).m_dfPe;
                }
            else if (bAverage)
                {
                // Average
                double dfPe = 0.;
                for (int nRow = nRowBeg; nRow < nRowEnd; nRow++)
                    dfPe += m_lstPePure.at(nRow).m_dfPe;
                oSample.m_dfPe = dfPe / (double)(nRowEnd - nRowBeg);
                }
            else
                {
                // Median
                double dfPe = 0.;
                QList<double> oList;
                for (int nRow = nRowBeg; nRow < nRowEnd; nRow++)
                    oList.append(m_lstPePure.at(nRow).m_dfPe);
                qSort(oList);
                dfPe = oList.at((nRowEnd - nRowBeg) / 2);
                oSample.m_dfPe = dfPe;
                }
            m_lstPeSmoothed.append(oSample);
            }
        }

    // then combine all raw samples for one PEC-point and smooth again
    // don't work on smoothed data to avoid problems with gaps and such ...
    QVector<QList<double> > aPeSorted;
    aPeSorted.resize(PecData::c_nRowNum);
    for (int nRow = 0; nRow < m_lstPePure.size(); nRow++)
        {
        double dfWormPos = m_lstPePure.at(nRow).m_dfWorm;
        double dfPe = m_lstPePure.at(nRow).m_dfPe;
        int nWormPos = (int)floor(dfWormPos);
        if ((nWormPos >= 0) && (nWormPos < PecData::c_nRowNum))
            aPeSorted[nWormPos].append(dfPe);
        }
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        QList<double> lstPe;
        int nSmoothMin = - nSmoothWidth / 2;
        int nSmoothMax = (nSmoothWidth + 1) / 2;
        if (nSmoothMax < 1)
            nSmoothMax = 1;
        for (int nSmooth = nSmoothMin ; nSmooth < nSmoothMax; nSmooth++)
            lstPe += aPeSorted[(nRow + nSmooth + PecData::c_nRowNum) % PecData::c_nRowNum];
        double dfPe = 0.;
        if (lstPe.isEmpty())
            {
            // currently we rely on the user choosing adequate smoothing
            // todo: interpolate?
            }
        else if (bAverage)
            {
            // Average
            for (int nCnt = 0; nCnt < lstPe.size(); nCnt++)
                dfPe += lstPe.at(nCnt);
            dfPe /= (double)(lstPe.size());
            }
        else
            {
            // Median
            qSort(lstPe);
            dfPe = lstPe.at((lstPe.size()) / 2);
            }
        m_aPeProcessed[nRow] = dfPe;
        }

    // if in Touchup-mode, add current PEC table
    // todo: rethink this with PecIdeal
    if (bTouchup)
        {
        for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
            m_aPeProcessed[nRow] += (double)(*pPecTable)[nRow].m_nPos * dfPecPosScale;
        }

    // zero out PecIdeal
    // (will be recalculated when generating PEC table, as it needs worm wheel teeth
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        m_aPecIdeal[nRow] = 0.;

    // m_lstPeriod is the list of periods.
    // each entry holds first and last sample of the period
    // (as index into the list of smoothed PE data)
    // and minimum and maximum PE value of period
    m_lstPeriod.clear();

    Period oPeriod;
    oPeriod.m_nRowBeg = -1;
    double dfWormOld = (double)PecData::c_nRowNum * 4.;
    for (int nRow = 0; nRow < m_lstPeSmoothed.size(); nRow++)
        {
        oPeriod.m_nRowEnd = nRow;
        double dfWormPos = m_lstPeSmoothed.at(nRow).m_dfWorm;
        // a leap of WormPos indicates end of current period
        if (fabs(dfWormPos - dfWormOld) > (double)PecData::c_nRowNum / 2.)
            {
            if (oPeriod.m_nRowBeg >= 0)
                m_lstPeriod.append(oPeriod);
            oPeriod.m_nRowBeg = nRow;
            oPeriod.m_dfPeMin = 1e10;
            oPeriod.m_dfPeMax = -1e10;
            }
        double dfPe = m_lstPeSmoothed.at(nRow).m_dfPe;
        if (oPeriod.m_dfPeMin > dfPe)
            oPeriod.m_dfPeMin = dfPe;
        if (oPeriod.m_dfPeMax < dfPe)
            oPeriod.m_dfPeMax = dfPe;
        dfWormOld = dfWormPos;
        }
    if (oPeriod.m_nRowBeg >= 0)
        {
        oPeriod.m_nRowEnd += 1;
        m_lstPeriod.append(oPeriod);
        }

    // convert processed PE data into a profile for compensating the PE
    // to do so, invert the absolute PE profile:
    // - add ideal worm vs. RA ratio to measured PE data
    // - swap axis
    // - subtract ideal RA vs. worm ratio


    // center PE data around 0, add nominal gear ratio and normalize into range 0 ... PecData::c_nRowNum
    double dfMin = 1000.;
    double dfMax = -1000.;
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        if (dfMin > m_aPeProcessed.at(nRow))
            dfMin = m_aPeProcessed.at(nRow);
        if (dfMax < m_aPeProcessed.at(nRow))
            dfMax = m_aPeProcessed.at(nRow);
        }
    double dfCenter = (dfMin + dfMax) / 2.;
    double dfArcSecsPerPecPoint = 360. * 3600. / (double)nWormWheelTeeth / (double)PecData::c_nRowNum;
    QVector<double> aAbsProfile(PecData::c_nRowNum + 1);
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        aAbsProfile[nRow] = (m_aPeProcessed.at(nRow) - dfCenter) / dfArcSecsPerPecPoint + (double)nRow;
    aAbsProfile[PecData::c_nRowNum] = aAbsProfile[0] + (double)PecData::c_nRowNum;

    // invert profile:
    // binary search for PEC points that are just below and above RA value,
    // then linearily interpolate between these two points
    QVector<double> aInvProfile(PecData::c_nRowNum);
    for (int nCol = 0; nCol < PecData::c_nRowNum; nCol++)
        {
        double dfColMid = (double)nCol;
        int nRowBot = 0;
        int nRowTop = PecData::c_nRowNum;
        while (nRowTop - nRowBot > 1)
            {
            int nRowMid = (nRowTop + nRowBot) / 2;
            if (dfColMid < aAbsProfile[nRowMid])
                nRowTop = nRowMid;
            else
                nRowBot = nRowMid;
            }
        double dfColBot = aAbsProfile[nRowBot];
        double dfColTop = aAbsProfile[nRowTop];
        aInvProfile[nCol] = (double)nRowBot;
        if ((dfColTop - dfColBot) > 1e-6)
            aInvProfile[nCol] += (dfColMid - dfColBot) / (dfColTop - dfColBot);
        }

    // normalize back to arcseconds and subtract nominal gear ratio
    // invert sign, as MCU firmware needs the PEC table flipped ...
    // undo centering around 0
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        m_aPecIdeal[nRow] = (aInvProfile.at(nRow) - (double)nRow) * dfArcSecsPerPecPoint * -1. + dfCenter;
    }


